
(declare (unit util-bst))

(import duck)

(module*
 util-bst
 #:doc ("Binary Search Tree implementation")
 (
  make-bst
  bst?

  bst-empty?
  ;;bst-ref
  ;;bst-contains?

  ;;bst-set
  ;;bst-remove ;;

  ;;bst->kvv
  ;;kvv->bst
  ;;kvv-filter
  
  ;;bst-balance

  util-bst-tests!
  )

 (import scheme
	 util-tag
	 testing)

 ;; Unique tags
 (define TAG-BST (make-tag bst))
 (define TAG-KVV (make-tag kvv))

 (define/doc (make-bst subtag EQ? <?)
   ("Creates empty BST with given comparators")
   (cons TAG-BST
	 (cons (cons #f 0)
	       (cons subtag
		     (cons EQ? <?)))))

 ;; BST accessors
 (define bst-tag car)
 (define bst-root+count cadr)
 (define bst-root caadr)
 (define bst-count cdadr)
 (define bst-type cddr)
 (define bst-subtag caddr)
 (define bst-comparators cdddr)
 (define bst-EQ? cadddr)
 (define bst-<? cddddr)

 (define/doc ((bst? subtag) v)
   ("Curried predicate for particular bst type.")
   (and (pair? v)
	(eq? (bst-tag v) TAG-BST)
	(eq? (bst-subtag v) subtag)))

 (define/doc (bst-empty? bst)
   ("Returns #t if given BST is empty.")
   (not (bst-root bst)))

 ;; Wrapper to setup comparators
 (define-syntax define-bst-proc
   (syntax-rules ()
     ((_ EQ? <? (name bst . args) expr ...)
      (define (name bst . args)
	(let ((EQ? (bst-eq? bst))
	      (<? (bst-<? bst)))
	  expr ...)))))

 ;; Module self-tests
 (define (util-bst-tests!)
   (run-tests
    util-bst
    (test-equal? make-bst
		 (make-bst 'fixnum eq? <)
		 `(,TAG-BST
		   . ((#f . 0)
		      . (fixnum . (,eq? . ,<)))))
    (test-true bst?
	       ((bst? 'fixnum) (make-bst 'fixnum eq? <)))
    (test-false bst?
		((bst? 'fixnum) (make-bst 'string eq? <)))
    (test-false bst?
		((bst? 'fixnum) "string"))
    (test-true bst-empty?
	       (bst-empty? (make-bst 'fixnum eq? <)))
    ))

 )

(import util-bst)
(util-bst-tests!)
